#ifndef SB_COMMON_H
#define SB_COMMON_H

#define DCTBASE 0				// DCT:1 DFT:0
#define MULTIGPU 1				// True:1 False:0
#define MATVECWITHCUSPARSE 0	// True:1 False:0
#define GPUGLOBALMEMORY 6		// stands for 6GB
#define THREADS 64				// number of threads

// resid = X * param - Y
// || resid ||_2^2 / 2 + l1 * || param ||_1 + l2 * || param - param1 ||_1
double getObjValue( cublasHandle_t handle, double *resid, double *param, double *param1, double lamb1, double lamb2, int N, int P, size_t sizeP1 )
{
	double val, temp;

	cublasDnrm2( handle, N,
		resid, 1, // const double *x, int incx,
		&val ); // double *result
	val = val * val / 2.0;

	cudaMemcpy( param1, param+1, sizeP1, DtoD );
	cublasDaxpy( handle, P-1,
		&MINUSONE, // const double *alpha,
		param, 1, // const double *x, int incx,
		param1, 1 ); // double *y, int incy

	cublasDasum( handle, P,
		param, 1, // const double *x, int incx,
		&temp ); // double *result
	val += lamb1 * temp;

	cublasDasum( handle, P-1,
		param1, 1, // const double *x, int incx,
		&temp ); // double *result
	val += lamb2 * temp;

	return val;
}

__global__ void matvec( double *res, double *param, double mu1, double mu2, int P, int M )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	
	if (IsValidRange)
	{
		double temp1 = 0, temp2 = 0, temp3 = 0, temp4 = 0;
		int k = t % M;
		int j = (int) (t / M);
		if (k > 0) temp1 = -1*mu2 * param[t-1];
		if (k < (M-1)) temp2 = -1*mu2 * param[t+1];
		if (j > 0) temp3 = -1*mu2 * param[t-M];
		if (j < (M-1)) temp4 = -1*mu2 * param[t+M];
		res[t] = (mu1+4*mu2) * param[t] + temp1 + temp2 + temp3 + temp4; // NOT concerning boundary conditions
		//res[t] = (mu1+4*mu2) * param[t] + temp1 + temp2;
	}
}

__global__ void SoftThr( double *x, double lamb, double mu, int P )
{
	int t = blockIdx.x * THREADS + threadIdx.x;
	double lam = lamb / mu;
	if (IsValidRange)
	{
		if (x[t] > lam)
			x[t] -= lam;
		else
		{
			if (x[t] < -lam)
				x[t] += lam;
			else
				x[t] = 0;
		}
	}
}

#endif /* SB_COMMON_H */