#ifndef SB_MACRO_H
#define SB_MACRO_H

#include <stdio.h>
#define ABS(a)			(((a) < 0) ? -(a) : (a))
#define HtoD			cudaMemcpyHostToDevice
#define DtoH			cudaMemcpyDeviceToHost
#define DtoD			cudaMemcpyDeviceToDevice
#define IsValidRange	(t < P)
#define IsLeftmost		(t == 0)
#define IsRightmost		(t == (P-1))
#define IsNOTLeftmost	(t > 0)
#define IsNOTRightmost	(t < (P-1))

#define myMalloc( a, b )									\
do {														\
	if ( cudaSuccess != cudaMalloc( (void**)&a, b ) ) {		\
		printf( "cudaMalloc Fail (" #a ")\n" );				\
		cublasDestroy(handle);								\
		return 0;											\
	}														\
} while (0)

#define cudaMemcpyMulti2Multi( a, b, c )					\
do {														\
	for (i=0; i<deviceCount; i++)							\
		cudaMemcpyAsync( a[i], b[i], c, DtoD );				\
	cudaEventRecord(stop);									\
	cudaEventSynchronize(stop);								\
} while (0)

#define cudaMemcpySingle2Multi( a, b, c )					\
do {														\
	for (i=0; i<deviceCount; i++)							\
		cudaMemcpyAsync( a[i], b, c, DtoD );				\
	cudaEventRecord(stop);									\
	cudaEventSynchronize(stop);								\
} while (0)

#define cudaMemcpyResidBlock															\
do {																					\
	for (i=0; i<deviceCount; i++)														\
		cudaMemcpyAsync( resid + i*tmpN, residblock[i], tmpN * sizeof(double), DtoD );	\
	cudaEventRecord(stop);																\
	cudaEventSynchronize(stop);															\
} while (0)

#define cublasMultiDaxpy( a, b, c, d, e, f, g )				\
do {														\
	for (i=0; i<deviceCount; i++)							\
		cublasDaxpy( a, b, &c, d[i], e, f, g );				\
} while (0)


#define cublasMultiDgemv( a, b, c, d, e, f, g, h, j, k, l, m )				\
do {																		\
	for (i=0; i<deviceCount; i++)											\
	{																		\
		cudaSetDevice( deviceList[i] );										\
		cublasDgemv( a[i], b, c, d, &e, f[i], g, h[i], j, &k, l[i], m );	\
	}																		\
	cudaSetDevice(0);														\
} while (0)

#define cublasMulti2Dgemv							\
do {												\
	for (i=0; i<deviceCount; i++)					\
	{												\
		cudaSetDevice( deviceList[i] );				\
		cublasDgemv( handles[i], CUBLAS_OP_T, P, tmpN, &ONE, Xblock[i], P, sizePblock[i], 1, &ZERO, residblock[i], 1 );	\
		cublasDgemv( handles[i], CUBLAS_OP_N, P, tmpN, &ONE, Xblock[i], P, residblock[i], 1, &ZERO, sizePblock[i], 1 );	\
	}												\
	cudaSetDevice(0);								\
} while (0)

#define DCTsolver( out, in )						\
do {												\
	cudaMemset( r_real_d, 0, size4P8M4 );\
	real2complex<<<nb, THREADS>>>( r_real_d, in, P, M, FFTrowsize, FFTindex[0], FFTindex[1], FFTindex[2], FFTindex[3] );\
	cufftExecD2Z( plan, r_real_d, r_complex_d );\
	cudaMemset( r_real_d, 0, size4P8M4 );\
	solve_poisson<<<nb, THREADS>>>( r_real_d, r_complex_d, P, M, FFTrowsize, FFTindex[0], FFTindex[1], FFTindex[2], FFTindex[3], kx_d );\
	cufftExecD2Z( plan, r_real_d, r_complex_d );\
	complex2real<<<nb, THREADS>>>( out, r_complex_d, P, M, FFTrowsize, FFTindex[0], scale );\
} while (0)

#define DFTsolver( out, in )													\
do {																			\
	cudaMemset( r_real_d, 0, 2 * sizeP );										\
	real2complex_dft<<<nb, THREADS>>>( r_real_d, in, P, M );					\
	cufftExecZ2Z( plan, r_real_d, r_complex_d, CUFFT_FORWARD );					\
	solve_poisson_dft<<<nb, THREADS>>>( r_real_d, r_complex_d, P, M, kx_d );	\
	cufftExecZ2Z( plan, r_real_d, r_complex_d, CUFFT_FORWARD );					\
	complex2real_dft<<<nb, THREADS>>>( out, r_complex_d, P, M );				\
} while (0)

#define DCTsolver_pcgls( out, in )					\
do {												\
	cudaMemset( r_real_d, 0, size4P8M4 );\
	real2complex<<<nb, THREADS>>>( r_real_d, in, P, M, FFTrowsize, FFTindex[0], FFTindex[1], FFTindex[2], FFTindex[3] );\
	cufftExecD2Z( plan, r_real_d, r_complex_d );\
	cudaMemset( r_real_d, 0, size4P8M4 );\
	solve_poisson_pcgls<<<nb, THREADS>>>( r_real_d, r_complex_d, P, M, FFTrowsize, FFTindex[0], FFTindex[1], FFTindex[2], FFTindex[3], kx_d );\
	cufftExecD2Z( plan, r_real_d, r_complex_d );\
	complex2real<<<nb, THREADS>>>( out, r_complex_d, P, M, FFTrowsize, FFTindex[0], scale );\
} while (0)

#define DFTsolver_pcgls_old( out, in )												\
do {																			\
	cudaMemset( r_real_d, 0, 2 * sizeP );										\
	real2complex_dft<<<nb, THREADS>>>( r_real_d, in, P, M );					\
	cufftExecZ2Z( plan, r_real_d, r_complex_d, CUFFT_FORWARD );					\
	solve_poisson_pcgls_dft<<<nb, THREADS>>>( r_real_d, r_complex_d, P, M, kx_d );	\
	cufftExecZ2Z( plan, r_real_d, r_complex_d, CUFFT_FORWARD );					\
	complex2real_dft<<<nb, THREADS>>>( out, r_complex_d, P, M );				\
} while (0)

#define DFTsolver_pcgls( out, in )												\
do {																			\
	cudaMemset( r_real_d, 0, 2 * sizeP );										\
	real2complex_dft<<<nb, THREADS>>>( r_real_d, in, P, M );					\
	cufftExecZ2Z( plan, r_real_d, r_real_d, CUFFT_FORWARD );					\
	solve_poisson_pcgls_dft<<<nb, THREADS>>>( r_real_d, P, M, kx_d );			\
	cufftExecZ2Z( plan, r_real_d, r_real_d, CUFFT_FORWARD );					\
	complex2real_dft<<<nb, THREADS>>>( out, r_real_d, P, M );					\
} while (0)

double MINUSONE		= -1.0;
double ONE			= 1.0;
double ZERO			= 0.0;

#endif /* SB_MACRO_H */