#include <stdio.h>
#include <stdlib.h>
#include <time.h>

double *SBFL(int N, int P, double *dataX, double *dataY, double lamb1, double lamb2, double mu1, double mu2, double epsil, double tol, double *obj, int *niter);
double *SBFLdenoising(int N, int P, double *dataY, double lamb1, double lamb2, double mu1, double mu2, double epsil, double tol, double *obj, int *niter);

int main(int argc, char *argv[])
{
	int N, P;
	double *X, *Y;
	double *solution;
	char fn1[100], fn2[100], fn3[100], fn4[100];
	
	int i, j;
	FILE *in1, *in2, *in3, *in4;
	bool verbose = false;
	clock_t start;

	// parameters
	int niter = 100;
	double lamb1, lamb2;
	double epsil = 1e-6;
	double tol = 1e-6;
	double obj = 0;

	if (argc < 2)
	{
		printf("Please input a folder name\n");
		return 1;
	}
	else
	{
		for (i=0; argv[1][i]!=0; fn1[i]=argv[1][i], fn2[i]=argv[1][i], fn3[i]=argv[1][i], fn4[i]=argv[1][i], i++);

		for (j=i; argv[2][j-i]!=0; fn1[j] = argv[2][j-i], fn2[j] = argv[2][j-i], fn3[j] = argv[2][j-i], fn4[j] = argv[2][j-i], j++);
		i = j; fn1[i++] = '/'; fn1[i++] = 'X'; fn1[i++] = 'b'; fn1[i++] = '.'; fn1[i++] = 't'; fn1[i++] = 'x'; fn1[i++] = 't'; fn1[i] = 0;
		i = j; fn2[i++] = '/'; fn2[i++] = 'y'; fn2[i++] = 'b'; fn2[i++] = '.'; fn2[i++] = 't'; fn2[i++] = 'x'; fn2[i++] = 't'; fn2[i] = 0;
		i = j; fn3[i++] = '/'; fn3[i++] = 'X'; fn3[i++] = '.'; fn3[i++] = 't'; fn3[i++] = 'x'; fn3[i++] = 't'; fn3[i] = 0;
		i = j; fn4[i++] = '/'; fn4[i++] = 'y'; fn4[i++] = '.'; fn4[i++] = 't'; fn4[i++] = 'x'; fn4[i++] = 't'; fn4[i] = 0;
		for (j=0; argv[2][j]!='x'; j++);
		argv[2][j] = 0;
		N = atoi(argv[2]);
		P = atoi(&argv[2][j+1]);
		
		lamb1 = atof(argv[3]);
		lamb2 = atof(argv[4]);
		if (argc > 5) epsil = atof(argv[5]);
		if (argc > 6) tol = atof(argv[6]);

		if (argc == 8)
			if (argv[7][1]=='v')
				verbose = true;
	}
	
	// initialization
	in1 = fopen(fn1, "rb");
	in2 = fopen(fn2, "rb");
	bool denoising = false;

	if (in1 != NULL)
	{
		X = (double *) malloc( N * P * sizeof(double) );
		Y = (double *) malloc( N * sizeof(double) );
		fread(X, sizeof(double), (size_t) N * P, in1);
		fread(Y, sizeof(double), (size_t) N, in2);
		fclose(in1);
		fclose(in2);
	}
	else
	{
		in3 = fopen(fn3, "r");
		in4 = fopen(fn4, "r");

		if (in3 != NULL)
		{
			X = (double *) malloc( N * P * sizeof(double) );
			Y = (double *) malloc( N * sizeof(double) );
			for (i=0; i<N; i++)
			{
				fscanf(in4, "%lf", Y+i);
				for (j=0; j<P; j++) fscanf(in3, "%lf", X+i*P+j);
			}
			fclose(in3);
			fclose(in4);
		}
		else // denoising case
		{
			denoising = true;
			Y = (double *) malloc( N * sizeof(double) );
			fread(Y, sizeof(double), (size_t) N, in2);
			fclose(in2);
		}
	}

	// MM
	start = clock();
	if (denoising)
		solution = SBFLdenoising(N, P, Y, lamb1, lamb2, 1, 1, epsil, tol, &obj, &niter);
	else
		solution = SBFL(N, P, X, Y, lamb1, lamb2, 1, 1, epsil, tol, &obj, &niter);
	if (solution)
	{
		printf("Time elapsed: %f\n", ((double)clock() - start) / CLOCKS_PER_SEC);
		printf("# of iterations: %d\n", niter);
		printf("Max differential: %.9lf\n" , obj);
	}
	else
	{
		fprintf(stderr, "Failed\n");
		return 1;
	}

	if (verbose)
	{
		printf("Solution:");
		if (P > 600)
		{
			printf(" (too many elements..)\n");
			for (j=0; j<10; j++) printf("%.3lf\t", solution[j]);
			printf("...\t");
			for (j=P-10; j<P; j++) printf("%.3lf\t", solution[j]);
		}
		else
		{
			printf("\n");
			for (j=0; j<P; j++)
				printf("%.3lf\t", solution[j]);
		}
		printf("\n");
	}
	
    free(solution);
	free(X);
	free(Y);

	return 0;
}
